#include "FindGeodesic.h"
#include <iostream>

class compare
{
public:
	bool operator() (std::pair<int, double>& p1, std::pair<int, double>& p2)
	{
		return p1.second > p2.second;
	}
};

Findgeodesic::Findgeodesic()
{

}

Findgeodesic::~Findgeodesic()
{

}

Findgeodesic::Findgeodesic(const Mesh & mesh_)
{
	my_mesh = mesh_;
	nv = my_mesh.n_vertices();
	dist.resize(nv);
	prev.resize(nv);
	one_ring.resize(nv);
	one_ring_length.resize(nv);

	int i = 0;
	for (auto v_it = my_mesh.vertices_begin(); v_it != my_mesh.vertices_end(); ++v_it)
	{
		int valence = my_mesh.valence(*v_it);
		std::vector<int> mid_one_ring(valence);
		std::vector<double> mid_one_ring_length(valence);
		int j = 0;
		for (auto halfedge = my_mesh.voh_iter(*v_it); halfedge.is_valid(); ++halfedge)
		{
			int halfedge_vertex = my_mesh.to_vertex_handle(*halfedge).idx();
			double length = my_mesh.calc_edge_length(*halfedge);
			mid_one_ring[j] = halfedge_vertex;
			mid_one_ring_length[j] = length;
			j++;
		}
		one_ring[i] = mid_one_ring;
		one_ring_length[i] = mid_one_ring_length;
		i++;
	}

	G.reserve(mesh_.n_vertices());
	for (const auto& vh : mesh_.vertices())
	{
		std::vector<std::pair<int, double>> vecpair;
		const auto& p0 = mesh_.point(vh);
		for (const auto& vvh : mesh_.vv_range(vh))
		{
			vecpair.push_back(std::pair<int, double>(vvh.idx(), (mesh_.point(vvh) - p0).norm()));
		}
		G.push_back(vecpair);
	}
}

void Findgeodesic::set_mesh(Mesh & mesh_)
{
	my_mesh = mesh_;
	nv = my_mesh.n_vertices();
	dist.resize(nv);
	prev.resize(nv);
	one_ring.resize(nv);
	int i = 0;
	for (auto v_it = my_mesh.vertices_begin(); v_it != my_mesh.vertices_end(); ++v_it)
	{
		int valence = my_mesh.valence(*v_it);
		std::vector<int> mid_one_ring(valence);
		int j = 0;
		for (auto halfedge = my_mesh.voh_iter(*v_it); halfedge.is_valid(); ++halfedge)
		{
			int halfedge_vertex = my_mesh.to_vertex_handle(*halfedge).idx();
			mid_one_ring[j] = halfedge_vertex;
			j++;
		}
		one_ring[i] = mid_one_ring;
		i++;
	}
}

void Findgeodesic::set_edges_weight(std::vector<std::vector<double>> & edges_wei)
{
	one_ring_length = edges_wei;

	int n = my_mesh.n_vertices();
	G.clear();
	G.reserve(n);
	for (auto vh : my_mesh.vertices())
	{
		std::vector<std::pair<int, double>> vecpair;
		unsigned v_idx = vh.idx();
		unsigned counter = 0;
		for (auto vvh : my_mesh.vv_range(vh))
		{
			vecpair.push_back(std::pair<int, double>(vvh.idx(), one_ring_length[v_idx][counter]));
			counter++;
		}
		G.push_back(vecpair);
	}
}

std::vector<int> Findgeodesic::Find_Geodesic(int sou_p, int tar_p)
{
	Dijkstra(sou_p, tar_p);
	distance_weight = dist[tar_p];
	/*std::cout <<"distance_weight: "<< distance_weight<< std::endl;
	std::vector<int> mid = searchPath(sou_p, tar_p);
	for (int i = 0; i < mid.size(); i++)
	{
	std::cout << "current shortest route: " << mid[i] << "  " << dist[mid[i]] << std::endl;
	}*/
	return searchPath(sou_p, tar_p);
}

std::vector<int> Findgeodesic::Find_Geodesic(int sou_p, std::vector<int>& is_target)
{
	int tar_p = Dijkstra(sou_p, is_target);
	/*std::vector<int> mid = searchPath(sou_p, tar_p);
	for (int i = 0; i < mid.size(); i++)
	{
		std::cout << "current shortest route: " << mid[i] << "  " << dist[mid[i]] << std::endl;
	}*/
	return searchPath(sou_p, tar_p);
}

void Findgeodesic::Find_Geodesic(int sou_p, std::vector<int>& is_target, int size_)
{
	Dijkstra(sou_p, is_target, size_);
}

std::vector<int> Findgeodesic::Change_Geodesic(int sou_p, int tar_p, std::vector<int> visit_or_not)
{
	Change_Dijkstra(sou_p, tar_p, visit_or_not);
	return searchPath(sou_p, tar_p);
}

void Findgeodesic::Change_Dijkstra(int sou_v, int tar_v, std::vector<int> visit_or_not)
{
	std::vector<int> s(nv);//жǷѾ·У0ʾûм룬1ʾ
	visit_or_not[tar_v] = -1;
	//ʼ
	for (int i = 0; i < nv; ++i)
	{
		dist[i] = maxint;
		s[i] = 0;
		prev[i] = -1;
	}
	//洢sou_v1-㵽õľ룬Ҷ1ǰĵ㶼sou_v
	/*for (int i = 0; i < one_ring[sou_v].size(); ++i)
	{
		dist[one_ring[sou_v][i]] = one_ring_length[sou_v][i];
		prev[one_ring[sou_v][i]] = sou_v;
	}*/
	dist[sou_v] = 0;//ʼ㵽ľΪ0
	s[sou_v] = 1;//ʼ·ļ

	std::priority_queue<std::pair<int, double>, std::vector<std::pair<int, double>>, compare> Q;
	//ȶ
	Q.push(std::pair<int, double>(sou_v, 0.0));
	while (!Q.empty())
	{
		//ҵԪ
		std::pair<int, double> top = Q.top();
		Q.pop();

		int i = top.first;
		s[i] = 1;
		double c = top.second;

		if (s[tar_v])
		{
			//std::cout << "current v: " << i << "  " << prev[i] << std::endl;
			return;
		}

		if (c <= dist[i])
		{
			//дsou_pı
			for (auto& pi : G[i])
			{
				int i2 = pi.first;
				double c2 = pi.second;
				if ((c2 + dist[i] < dist[i2]) && (visit_or_not[i2] == -1)) // update
				{
					dist[i2] = c2 + dist[i];
					prev[i2] = i;
					Q.push(std::pair<int, double>(i2, dist[i2]));
				}
			}
		}
	}

	//for (int i = 1; i < nv; ++i)
	//{
	//	double tmp = maxint;
	//	int u = sou_v;
	//	for (int j = 0; j < nv; ++j)
	//	{
	//		if ((!s[j]) && dist[j] < tmp && (visit_or_not[j] == -1))
	//		{
	//			u = j;//ǰ1-оõδʵĵ
	//			tmp = dist[j];
	//		}
	//	}
	//	s[u] = 1;
	//	visit_or_not[u] = 1;
	//	//dist
	//	for (int j = 0; j < one_ring[u].size(); ++j)
	//	{
	//		if (!s[one_ring[u][j]] && (visit_or_not[one_ring[u][j]] == -1))
	//		{
	//			double newdist = dist[u] + one_ring_length[u][j];
	//			if (newdist < dist[one_ring[u][j]])
	//			{
	//				dist[one_ring[u][j]] = newdist;
	//				prev[one_ring[u][j]] = u;
	//			}
	//		}
	//	}
	//	if (s[tar_v])
	//	{
	//		return;
	//	}
	//}
}

void Findgeodesic::Dijkstra(int sou_v, int tar_v)
{
	std::vector<int> s(nv);//жǷѾ·У0ʾûм룬1ʾ
	//ʼ
	for (int i = 0; i < nv; ++i)
	{
		dist[i] = maxint;
		s[i] = 0;
		prev[i] = -1;
	}
	//洢sou_v1-㵽õľ룬Ҷ1ǰĵ㶼sou_v
	/*for (int i = 0; i < one_ring[sou_v].size(); ++i)
	{
		dist[one_ring[sou_v][i]] = one_ring_length[sou_v][i];
		prev[one_ring[sou_v][i]] = sou_v;
	}*/
	dist[sou_v] = 0;//ʼ㵽ľΪ0
	s[sou_v] = 1;//ʼ·ļ

	std::priority_queue<std::pair<int, double>, std::vector<std::pair<int, double>>, compare> Q;
	//ȶ
	Q.push(std::pair<int, double>(sou_v, 0.0));
	while (!Q.empty())
	{
		//ҵԪ
		std::pair<int, double> top = Q.top();
		Q.pop();

		int i = top.first;
		s[i] = 1;
		double c = top.second;

		if (s[tar_v])
		{
			return;
		}

		if (c <= dist[i])
		{
			//дsou_pı
			for (auto& pi : G[i])
			{
				int i2 = pi.first;
				double c2 = pi.second;
				if (c2 + dist[i] < dist[i2]) // update
				{
					dist[i2] = c2 + dist[i];
					prev[i2] = i;
					Q.push(std::pair<int, double>(i2, dist[i2]));
				}
			}
		}
	}

	//for (int i = 1; i < nv; ++i)
	//{
	//	double tmp = maxint;
	//	int u = sou_v;
	//	for (int j = 0; j < nv; ++j)
	//	{
	//		if ((!s[j]) && dist[j] < tmp)
	//		{
	//			u = j;//ǰ1-оõδʵĵ
	//			tmp = dist[j];
	//		}
	//	}

	//	s[u] = 1;
	//	//dist
	//	for (int j = 0; j < one_ring[u].size(); ++j)
	//	{
	//		if (!s[one_ring[u][j]])
	//		{
	//			double newdist = dist[u] + one_ring_length[u][j];
	//			if (newdist < dist[one_ring[u][j]])
	//			{
	//				dist[one_ring[u][j]] = newdist;
	//				prev[one_ring[u][j]] = u;
	//			}
	//		}
	//	}
	//	if (s[tar_v])
	//	{
	//		return;
	//	}
	//}
}

void Findgeodesic::Dijkstra(int sou_v, std::vector<int>& is_target, int size_)
{
	int counter = 0;
	std::vector<int> s(nv);//жǷѾ·У0ʾûм룬1ʾ
	//ʼ
	for (int i = 0; i < nv; ++i)
	{
		dist[i] = maxint;
		s[i] = 0;
		prev[i] = -1;
	}
	//洢sou_v1-㵽õľ룬Ҷ1ǰĵ㶼sou_v
	/*for (int i = 0; i < one_ring[sou_v].size(); ++i)
	{
		dist[one_ring[sou_v][i]] = one_ring_length[sou_v][i];
		prev[one_ring[sou_v][i]] = sou_v;
	}*/
	dist[sou_v] = 0;//ʼ㵽ľΪ0
	//s[sou_v] = 1;//ʼ·ļ

	std::priority_queue<std::pair<int, double>, std::vector<std::pair<int, double>>, compare> Q;
	//ȶ
	Q.push(std::pair<int, double>(sou_v, 0.0));
	while (!Q.empty())
	{
		//ҵԪ
		std::pair<int, double> top = Q.top();
		Q.pop();
		int i = top.first;
		double c = top.second;
		while (s[i])
		{
			top = Q.top();
			Q.pop();
			i = top.first;
			c = top.second;
		}
		s[i] = 1;
		if (is_target[i])
		{
			//std::cout << "current v: " << i <<"  "<<prev[i]<<"  "<<size_<< std::endl;
			counter++;
			if (counter == size_)
			{
				return;
			}
		}

		if (c <= dist[i])
		{
			//дsou_pı
			for (auto& pi : G[i])
			{
				int i2 = pi.first;
				double c2 = pi.second;
				if ((c2 + dist[i] < dist[i2])&& (!s[i2])) // update
				{
					dist[i2] = c2 + dist[i];
					prev[i2] = i;
					Q.push(std::pair<int, double>(i2, dist[i2]));
				}
			}
		}
	}

	//for (int i = 1; i < nv; ++i)
	//{
	//	double tmp = maxint;
	//	int u = sou_v;
	//	for (int j = 0; j < nv; ++j)
	//	{
	//		if ((!s[j]) && dist[j] < tmp)
	//		{
	//			u = j;//ǰ1-оõδʵĵ
	//			tmp = dist[j];
	//		}
	//	}
	//	s[u] = 1;
	//	//dist
	//	for (int j = 0; j < one_ring[u].size(); ++j)
	//	{
	//		if (!s[one_ring[u][j]])
	//		{
	//			double newdist = dist[u] + one_ring_length[u][j];
	//			if (newdist < dist[one_ring[u][j]])
	//			{
	//				dist[one_ring[u][j]] = newdist;
	//				prev[one_ring[u][j]] = u;
	//			}
	//		}
	//	}
	//	if (is_target[u])
	//	{
	//		counter++;
	//		if (counter == size_)
	//		{
	//			return;
	//		}
	//	}
	//}
}

int Findgeodesic::Dijkstra(int sou_v, std::vector<int> is_target)
{
	std::vector<int> s(nv);//жǷѾ·У0ʾûм룬1ʾ
	//ʼ
	for (int i = 0; i < nv; ++i)
	{
		dist[i] = maxint;
		s[i] = 0;
		prev[i] = -1;
	}
	//洢sou_v1-㵽õľ룬Ҷ1ǰĵ㶼sou_v
	/*for (int i = 0; i < one_ring[sou_v].size(); ++i)
	{
	dist[one_ring[sou_v][i]] = one_ring_length[sou_v][i];
	prev[one_ring[sou_v][i]] = sou_v;
	}*/
	dist[sou_v] = 0;//ʼ㵽ľΪ0
	s[sou_v] = 1;//ʼ·ļ

	//std::priority_queue<std::pair<int, double>, std::vector<std::pair<int, double>>, compare> Q;
	////ȶ
	//Q.push(std::pair<int, double>(sou_v, 0.0));
	//while (!Q.empty())
	//{
	//	//ҵԪ
	//	std::pair<int, double> top = Q.top();
	//	Q.pop();

	//	int i = top.first;
	//	double c = top.second;
	//	s[i] = 1;

	//	if (is_target[i])
	//	{
	//		return i;
	//	}

	//	if (c <= dist[i])
	//	{
	//		//дsou_pı
	//		for (auto& pi : G[i])
	//		{
	//			int i2 = pi.first;
	//			double c2 = pi.second;
	//			if (c2 + dist[i] < dist[i2]) // update
	//			{
	//				dist[i2] = c2 + dist[i];
	//				prev[i2] = i;
	//				Q.push(std::pair<int, double>(i2, dist[i2]));
	//			}
	//		}
	//	}
	//}

	for (int i = 1; i < nv; ++i)
	{
		double tmp = maxint;
		int u = sou_v;
		for (int j = 0; j < nv; ++j)
		{
			if ((!s[j]) && dist[j] < tmp)
			{
				u = j;//ǰ1-оõδʵĵ
				tmp = dist[j];
			}
		}
		s[u] = 1;
		//dist
		for (int j = 0; j < one_ring[u].size(); ++j)
		{
			if (!s[one_ring[u][j]])
			{
				double newdist = dist[u] + one_ring_length[u][j];
				if (newdist < dist[one_ring[u][j]])
				{
					dist[one_ring[u][j]] = newdist;
					prev[one_ring[u][j]] = u;
				}
			}
		}
		if (is_target[u])
		{
			return u;
		}
	}
}

std::vector<int> Findgeodesic::searchPath(int sou_v, int tar_v)
{
	if (sou_v == tar_v)
	{
		std::cout << "sou_v == tar_v" << std::endl;
		std::vector<int> geo;
		return geo;
	}
	std::vector<int> que(dist.size());
	int tot = 0;
	que[tot] = tar_v;
	tot++;
	int tmp = prev[tar_v];
	while (tmp != sou_v)
	{
		que[tot] = tmp;
		tot++;
		if (tot == dist.size() || tmp == -1)
		{
			std::cout << "overflow !"<< std::endl;
		}
		tmp = prev[tmp];
	}
	que[tot] = sou_v;
	std::vector<int> geo;
	for (int i = 0; i <= tot; ++i)
	{
		geo.push_back(que[tot - i]);
	}
	return geo;
}